# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Any, Dict, Optional

import tensorflow as tf

import nncf
from nncf.common.quantization.structs import QuantizationScheme as QuantizationMode
from nncf.common.quantization.structs import QuantizerConfig
from nncf.common.quantization.structs import QuantizerSpec
from nncf.tensorflow.layers.custom_objects import NNCF_CUSTOM_OBJECTS
from nncf.tensorflow.layers.custom_objects import NNCF_QUANTIZATION_OPERATIONS
from nncf.tensorflow.layers.data_layout import get_channel_axis
from nncf.tensorflow.layers.data_layout import get_channel_size
from nncf.tensorflow.layers.operation import NNCFOperation
from nncf.tensorflow.quantization.functions import asymmetric_quantize
from nncf.tensorflow.quantization.functions import symmetric_quantize


class TFQuantizerSpec(QuantizerSpec):
    def __init__(
        self,
        num_bits: int,
        mode: QuantizationMode,
        signedness_to_force: Optional[bool],
        narrow_range: bool,
        half_range: bool,
        per_channel: bool,
    ):
        super().__init__(num_bits, mode, signedness_to_force, narrow_range, half_range)
        self.per_channel = per_channel

    @classmethod
    def from_config(cls, qconfig: QuantizerConfig, narrow_range: bool, half_range: bool) -> "TFQuantizerSpec":
        return cls(
            qconfig.num_bits, qconfig.mode, qconfig.signedness_to_force, narrow_range, half_range, qconfig.per_channel
        )

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            "num_bits": self.num_bits,
            "mode": self.mode,
            "signedness_to_force": self.signedness_to_force,
            "narrow_range": self.narrow_range,
            "half_range": self.half_range,
            "per_channel": self.per_channel,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TFQuantizerSpec":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)


class Quantizer(NNCFOperation):
    """
    Base class for all NNCF quantization operations.
    """

    def __init__(self, name: str):
        """
        Initializes internal NNCF quantization operation state.

        :param name: Unique operation name in algorithm scope.
        """
        super().__init__(name)
        self.enabled = True
        self._eps = 1e-16
        self._pre_processing_fn = self._make_pre_processing_fn()
        self._post_processing_fn = self._make_post_processing_fn()

    @property
    def mode(self) -> str:
        """
        Returns mode of the quantization (symmetric or asymmetric).

        :return: The mode of the quantization.
        """
        raise NotImplementedError

    def call(self, inputs, weights, training):
        """
        The method applies quantization to the input tensor if the quantizer is enabled,
        otherwise, if the quantizer is disabled, the method returns the input tensor as is.

        :param inputs: Input tensor.
        :param weights: Quantizer's weights.
        :param training: True if operation called in training mode else False
        :return: Output tensor.
        """
        if not self.enabled:
            return inputs
        transformed = self._pre_processing_fn(inputs)
        quantized = self.quantize(transformed, weights, training)
        outputs = self._post_processing_fn(quantized)
        return outputs

    def quantize(self, inputs, weights, training):
        """
        Apply quantization to the input tensor.

        :param inputs: Input tensor.
        :param weights: Quantizer's weights.
        :param training: True if operation called in training mode else False
        :return: Quantized tensor.
        """
        raise NotImplementedError

    def apply_range_initialization(self, weights, min_values, max_values, min_range=0.1, eps=0.01):
        """
        Initialize quantizer parameters using minimum and maximum weight values.

        :param weights: Quantizer's weights.
        :param min_values: Minimum weight values.
        :param max_values: Maximum weight values.
        :param min_range: Minimum range.
        :param eps: Smoothing coefficient for ranges: min_range = maximum(min_range, eps * max_range).
        """
        raise NotImplementedError

    def setup_input_transformation(self, input_shape, channel_axes):
        """
        Setup input transformation that the per-channel quantization can be applied to input tensor.
        The TensorFlow fake_quant_with_min_max_vars_per_channel supports only inputs tensor one of
        the shapes: [d], [b, d] [b, h, w, d]. For this reason, Quantizer transforms any inputs tensor
        to one of the supported shapes, then quantizes and then transforms quantized tensor to
        the original inputs shape.

        :param input_shape: Shape of the input.
        :param channel_axes: Channel axes.
        """
        try:
            self._pre_processing_fn, self._post_processing_fn = Quantizer._make_transformation_fns(
                input_shape, channel_axes
            )
        except NotImplementedError as e:
            raise NotImplementedError(f"Additional information: quantizer name {self.name}") from e

    @staticmethod
    def _make_transformation_fns(input_shape, channel_axes):
        fns_registry = []
        if isinstance(channel_axes, (tuple, list)):
            switch_counter = 0
            accumulate = False
            new_shape = []
            new_channel_axes = None
            for axis, val in enumerate(input_shape):
                if axis in channel_axes:
                    if accumulate:
                        new_shape[-1] *= val
                    else:
                        accumulate = True
                        new_channel_axes = len(new_shape)
                        new_shape.append(val)
                        switch_counter += 1
                else:
                    accumulate = False
                    new_shape.append(val)
            if switch_counter > 1:
                raise NotImplementedError(
                    "Quntizer could not transform input to apply per-channel quantization: "
                    f"input_shape {input_shape}, channel_axes {channel_axes}"
                )
            forward_params = {"shape": new_shape}
            backward_params = {"shape": input_shape}
            fns_registry.append((tf.reshape, forward_params, backward_params))
            input_shape = new_shape
            channel_axes = new_channel_axes

        ndims = len(input_shape)
        if channel_axes % ndims != ndims - 1:
            perm = [i for i, _ in enumerate(input_shape)]
            perm[channel_axes], perm[-1] = perm[-1], perm[channel_axes]
            params = {"perm": perm}
            fns_registry.append((tf.transpose, params, params))
            new_shape = list(input_shape)
            new_shape[channel_axes], new_shape[-1] = new_shape[-1], new_shape[channel_axes]
            input_shape = new_shape

        if ndims not in [1, 2, 4]:
            size = 1
            for val in input_shape[:-1]:
                size *= val
            forward_params = {"shape": [size, input_shape[-1]]}
            backward_params = {"shape": input_shape}
            fns_registry.append((tf.reshape, forward_params, backward_params))

        def fuse_functions(fns_registry):
            if not fns_registry:
                return fns_registry

            fused_fns_registry = []
            fn1 = fns_registry[0]
            for fn2 in fns_registry[1:]:
                if fn1[0] == fn2[0] == tf.reshape:
                    fn1 = (tf.reshape, fn2[1], fn1[2])
                else:
                    fused_fns_registry.append(fn1)
                    fn1 = fn2
            fused_fns_registry.append(fn1)
            return fused_fns_registry

        fused_fns_registry = fuse_functions(fns_registry)
        pre_processing_fn = Quantizer._make_pre_processing_fn(fused_fns_registry)
        post_processing_fn = Quantizer._make_post_processing_fn(fused_fns_registry)

        return pre_processing_fn, post_processing_fn

    @staticmethod
    def _make_pre_processing_fn(fns_registry=None):
        fns_list = []
        if fns_registry is None:
            fns_registry = []
        for fn in fns_registry:
            fns_list.append(partial(fn[0], **fn[1]))

        def pre_processing_fn(inputs):
            result = inputs
            for func in fns_list:
                result = func(result)
            return result

        return pre_processing_fn

    @staticmethod
    def _make_post_processing_fn(fns_registry=None):
        fns_list = []
        if fns_registry is None:
            fns_registry = []
        for fn in reversed(fns_registry):
            fns_list.append(partial(fn[0], **fn[2]))

        def post_processing_fn(inputs):
            result = inputs
            for func in fns_list:
                result = func(result)
            return result

        return post_processing_fn

    @staticmethod
    def _min_adj(bits, low, range_len, narrow_range):
        quants_count = 2**bits - (2 if narrow_range else 1)
        return range_len / quants_count * tf.round(quants_count * low / range_len)

    def get_quantizer_config(self) -> QuantizerConfig:
        """
        Used to get a current quantizer state in terms of QuantizerConfig objects.

        :return: A QuantizerConfig struct that corresponds to current state of the quantizer.
        """
        raise NotImplementedError

    def get_config(self):
        raise NotImplementedError


@NNCF_CUSTOM_OBJECTS.register()
@NNCF_QUANTIZATION_OPERATIONS.register(QuantizationMode.SYMMETRIC)
class SymmetricQuantizer(Quantizer):
    def __init__(self, name: str, qspec: TFQuantizerSpec):
        super().__init__(name)
        self.num_bits = qspec.num_bits
        self.per_channel = qspec.per_channel
        self.narrow_range = qspec.narrow_range
        self.signedness_to_force = qspec.signedness_to_force
        self._half_range = qspec.half_range

    @property
    def half_range(self):
        return self._half_range

    @property
    def mode(self) -> str:
        return QuantizationMode.SYMMETRIC

    def signed(self, op_weights) -> bool:
        """
        Returns `True` for signed quantization, `False` for unsigned.

        :return: `True` for signed quantization, `False` for unsigned.
        """
        signed_var = op_weights["signed_var"]
        return signed_var.numpy() < 0.0

    def build(self, input_shape, input_type, name, layer):
        channel_axes = None
        if self.per_channel:
            channel_axes = get_channel_axis(input_type, name, layer)
        return self._create_variables(layer, input_shape, channel_axes, name)

    def _create_variables(self, layer, input_shape, channel_axes, name: str = ""):
        shape = None
        if self.per_channel:
            self.setup_input_transformation(input_shape, channel_axes)
            shape = (get_channel_size(input_shape, channel_axes),)

        scale = layer.add_weight(
            name + "_scale", shape=shape, initializer=tf.keras.initializers.Constant(1.0), trainable=True
        )
        signed = layer.add_weight(
            name + "_signed",
            initializer=tf.keras.initializers.Constant(-1.0 if self.signedness_to_force in (True, None) else 0.0),
            trainable=False,
        )
        return {"scale_var": scale, "signed_var": signed}

    def apply_overflow_fix(self, weights):
        if self.num_bits != 8 or not self._half_range:
            raise nncf.InternalError(
                "Attempt to apply overflow issue fix to quantizer which is not configured for that."
            )

        # Multiplier to expand scale from 7 bit to 8 bit
        multiplier = 127 / 63 if self.narrow_range else 255 / 127
        weights["scale_var"].assign(multiplier * weights["scale_var"])
        self._eps *= multiplier
        self._half_range = False

    def quantize(self, inputs, weights, _):
        def _half_range_quantize():
            return symmetric_quantize(
                inputs,
                weights["scale_var"],
                weights["signed_var"],
                num_bits=self.num_bits - 1,
                per_channel=self.per_channel,
                narrow_range=self.narrow_range,
                eps=self._eps,
            )

        def _default_quantize():
            return symmetric_quantize(
                inputs,
                weights["scale_var"],
                weights["signed_var"],
                num_bits=self.num_bits,
                per_channel=self.per_channel,
                narrow_range=self.narrow_range,
                eps=self._eps,
            )

        if self._half_range:
            return _half_range_quantize()

        return _default_quantize()

    def apply_range_initialization(self, weights, min_values, max_values, min_range=0.1, eps=0.01):
        if self.signedness_to_force is None:
            sign = tf.reduce_any(tf.less(min_values, 0))
            weights["signed_var"].assign(-1.0 if sign else 0.0)
        ranges = tf.maximum(tf.abs(max_values), tf.abs(min_values))
        max_range = tf.reduce_max(ranges)
        lower_threshold = tf.maximum(eps * max_range, min_range)
        scale = tf.maximum(ranges, lower_threshold)
        weights["scale_var"].assign(scale)

    def get_quantizer_config(self) -> QuantizerConfig:
        return QuantizerConfig(
            num_bits=self.num_bits,
            mode=QuantizationMode.SYMMETRIC,
            signedness_to_force=self.signedness_to_force,
            per_channel=self.per_channel,
        )

    def get_config(self):
        qspec_dict = {
            "num_bits": self.num_bits,
            "mode": QuantizationMode.SYMMETRIC,
            "signedness_to_force": self.signedness_to_force,
            "narrow_range": self.narrow_range,
            "half_range": self._half_range,
            "per_channel": self.per_channel,
        }
        config = {
            "quantizer_spec": qspec_dict,
            "name": self.name,
        }
        return config

    @classmethod
    def from_config(cls, config):
        qspec_dict = config["quantizer_spec"]
        qspec = TFQuantizerSpec(
            num_bits=qspec_dict["num_bits"],
            mode=QuantizationMode.SYMMETRIC,
            signedness_to_force=qspec_dict["signedness_to_force"],
            narrow_range=qspec_dict["narrow_range"],
            half_range=qspec_dict["half_range"],
            per_channel=qspec_dict["per_channel"],
        )
        name = config["name"]
        return cls(name, qspec)


@NNCF_CUSTOM_OBJECTS.register()
@NNCF_QUANTIZATION_OPERATIONS.register(QuantizationMode.ASYMMETRIC)
class AsymmetricQuantizer(Quantizer):
    def __init__(self, name: str, qspec: TFQuantizerSpec):
        super().__init__(name)
        self.num_bits = qspec.num_bits
        self.narrow_range = qspec.narrow_range
        self.per_channel = qspec.per_channel
        self._half_range = qspec.half_range

    @property
    def half_range(self):
        return self._half_range

    @property
    def mode(self) -> str:
        return QuantizationMode.ASYMMETRIC

    def build(self, input_shape, input_type, name, layer):
        channel_axes = None
        if self.per_channel:
            channel_axes = get_channel_axis(input_type, name, layer)
        return self._create_variables(layer, input_shape, channel_axes, name)

    def _create_variables(self, layer, input_shape, channel_axes, name: str = ""):
        shape = None
        if self.per_channel:
            self.setup_input_transformation(input_shape, channel_axes)
            shape = (get_channel_size(input_shape, channel_axes),)

        input_low = layer.add_weight(
            name + "_input_low", shape=shape, initializer=tf.keras.initializers.Constant(0.0), trainable=True
        )
        input_range = layer.add_weight(
            name + "_input_range", shape=shape, initializer=tf.keras.initializers.Constant(1.0), trainable=True
        )
        return {"input_low_var": input_low, "input_range_var": input_range}

    def apply_overflow_fix(self, weights):
        if self.num_bits != 8 or not self._half_range:
            raise nncf.InternalError(
                "Attempt to apply overflow issue fix to quantizer which is not configured for that."
            )

        # Low value shift to expand quantize range from 7 bit to 8 bit properly
        weights["input_low_var"].assign(
            weights["input_low_var"]
            + self._min_adj(7, weights["input_low_var"], weights["input_range_var"] + self._eps, self.narrow_range)
        )
        # Multiplier to expand scale from 7 bit to 8 bit
        multiplier = 127 / 63 if self.narrow_range else 255 / 127
        weights["input_range_var"].assign(multiplier * weights["input_range_var"])
        self._eps *= multiplier
        self._half_range = False

    def quantize(self, inputs, weights, _):
        def _half_range_quantize():
            return asymmetric_quantize(
                inputs,
                weights["input_low_var"],
                weights["input_range_var"],
                num_bits=self.num_bits - 1,
                per_channel=self.per_channel,
                narrow_range=self.narrow_range,
                eps=self._eps,
            )

        def _default_quantize():
            return asymmetric_quantize(
                inputs,
                weights["input_low_var"],
                weights["input_range_var"],
                num_bits=self.num_bits,
                per_channel=self.per_channel,
                narrow_range=self.narrow_range,
                eps=self._eps,
            )

        if self._half_range:
            return _half_range_quantize()

        return _default_quantize()

    def apply_range_initialization(self, weights, min_values, max_values, min_range=0.1, eps=0.01):
        min_values = tf.minimum(min_values, 0)
        max_values = tf.maximum(max_values, 0)
        ranges = max_values - min_values
        max_range = tf.reduce_max(ranges)
        lower_threshold = tf.maximum(eps * max_range, min_range)
        correction = (tf.maximum(ranges, lower_threshold) - ranges) * 0.5
        input_low = min_values - correction
        input_range = ranges + 2 * correction
        weights["input_low_var"].assign(input_low)
        weights["input_range_var"].assign(input_range)

    def get_quantizer_config(self) -> QuantizerConfig:
        return QuantizerConfig(
            num_bits=self.num_bits,
            mode=QuantizationMode.ASYMMETRIC,
            signedness_to_force=None,
            per_channel=self.per_channel,
        )

    def get_config(self):
        qspec_dict = {
            "num_bits": self.num_bits,
            "mode": QuantizationMode.ASYMMETRIC,
            "signedness_to_force": None,
            "narrow_range": self.narrow_range,
            "half_range": self._half_range,
            "per_channel": self.per_channel,
        }
        config = {
            "quantizer_spec": qspec_dict,
            "name": self.name,
        }
        return config

    @classmethod
    def from_config(cls, config):
        qspec_dict = config["quantizer_spec"]
        qspec = TFQuantizerSpec(
            num_bits=qspec_dict["num_bits"],
            mode=QuantizationMode.ASYMMETRIC,
            signedness_to_force=None,
            narrow_range=qspec_dict["narrow_range"],
            half_range=qspec_dict["half_range"],
            per_channel=qspec_dict["per_channel"],
        )
        name = config["name"]
        return cls(name, qspec)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Type

import tensorflow as tf

from nncf.api.compression import CompressionAlgorithmController
from nncf.common.compression import NO_COMPRESSION_ALGORITHM_NAME
from nncf.common.compression import BaseCompressionAlgorithmController
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.logging import nncf_logger
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.backend import copy_model
from nncf.common.utils.registry import Registry
from nncf.tensorflow.api.compression import TFCompressionAlgorithmBuilder
from nncf.tensorflow.loss import TFZeroCompressionLoss

TF_COMPRESSION_ALGORITHMS = Registry("compression algorithm", add_name_as_attr=True)


@TF_COMPRESSION_ALGORITHMS.register(NO_COMPRESSION_ALGORITHM_NAME)
class NoCompressionAlgorithmBuilder(TFCompressionAlgorithmBuilder):
    def get_transformation_layout(self, _) -> TransformationLayout:
        return TransformationLayout()

    def _build_controller(self, model: tf.keras.Model) -> CompressionAlgorithmController:
        return NoCompressionAlgorithmController(model)

    def initialize(self, model: tf.keras.Model) -> None:
        pass

    def _get_algo_specific_config_section(self) -> Dict:
        return {}


class NoCompressionAlgorithmController(BaseCompressionAlgorithmController):
    def __init__(self, target_model: tf.keras.Model):
        super().__init__(target_model)
        self._loss = TFZeroCompressionLoss()
        self._scheduler = StubCompressionScheduler()

    @property
    def loss(self) -> TFZeroCompressionLoss:
        return self._loss

    @property
    def scheduler(self) -> StubCompressionScheduler:
        return self._scheduler

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        return NNCFStatistics()

    def strip(self, do_copy: bool = True) -> tf.keras.Model:
        model = self.model
        if do_copy:
            model = copy_model(self.model)
        return model


def get_compression_algorithm_builder(algo_name: str) -> Type[TFCompressionAlgorithmBuilder]:
    nncf_logger.info(f"Creating compression algorithm: {algo_name}")
    return TF_COMPRESSION_ALGORITHMS.get(algo_name)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Type

import torch

from nncf.common.graph import NNCFGraph
from nncf.common.logging import nncf_logger
from nncf.common.pruning.mask_propagation import MaskPropagationAlgorithm
from nncf.common.pruning.tensor_processor import NNCFPruningBaseTensorProcessor
from nncf.common.pruning.utils import PruningOperationsMetatypeRegistry
from nncf.torch.nncf_network import NNCFNetwork


class FilterReorderingAlgorithm(MaskPropagationAlgorithm):
    """
    Reorders filters based on reordering indexes encoded in the `output_mask` attribute in the nodes of
    model graph.
    """

    def __init__(
        self,
        model: NNCFNetwork,
        graph: NNCFGraph,
        pruning_operator_metatypes: PruningOperationsMetatypeRegistry,
        tensor_processor: Optional[Type[NNCFPruningBaseTensorProcessor]] = None,
    ):
        super().__init__(graph, pruning_operator_metatypes, tensor_processor)
        self._model = model

    def apply_reordering_indexes(self) -> None:
        """
        Applying propagated masks (which encodes indexes to reorder filters) for all nodes in topological order:
        1. running input_reorder method for this node
        2. running output_reorder method for this node
        """
        pruned_node_modules = []
        with torch.no_grad():
            for node in self._graph.topological_sort():
                node_cls = self.get_meta_operation_by_type_name(node.node_type)
                node_module = self._model.nncf.get_containing_module(node.node_name)
                if node_module not in pruned_node_modules:
                    node_cls.input_reorder(self._model, node, self._graph)
                    node_cls.output_reorder(self._model, node, self._graph)
                    pruned_node_modules.append(node_module)
            nncf_logger.debug("Finished mask applying step")

    def reorder_filters(self) -> None:
        """
        Model pruner work in two stages:
        1. Mask propagation: propagate pruning masks through the graph.
        2. Applying calculated masks
        """
        nncf_logger.info("Start reordering filters")
        self.mask_propagation()
        self.apply_reordering_indexes()
        nncf_logger.info("Finished reordering filters")



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import

import random
from collections import deque
from collections import namedtuple

import numpy as np

import nncf
from nncf import nncf_logger

# [reference] https://github.com/matthiasplappert/keras-rl/blob/master/rl/memory.py

# This is to be understood as a transition: Given `state0`, performing `action`
# yields `reward` and results in `state1`, which might be `terminal`.
Experience = namedtuple("Experience", "state0, action, reward, state1, terminal1")


def sample_batch_indexes(low, high, size):
    if high - low >= size:
        # We have enough data. Draw without replacement, that is each index is unique in the
        # batch. We cannot use `np.random.choice` here because it is horribly inefficient as
        # the memory grows. See https://github.com/numpy/numpy/issues/2764 for a discussion.
        # `random.sample` does the same thing (drawing without replacement) and is way faster.
        # try:
        #     r = xrange(low, high)
        # except NameError:
        r = range(low, high)
        batch_idxs = random.sample(r, size)
    else:
        # Not enough data. Help ourselves with sampling from the range, but the same index
        # can occur multiple times. This is not good and should be avoided by picking a
        # large enough warm-up phase.
        nncf_logger.warning(
            "Not enough entries to sample without replacement. Consider increasing your warm-up "
            "phase to avoid oversampling!"
        )
        batch_idxs = np.random.random_integers(low, high - 1, size=size)
    assert len(batch_idxs) == size
    return batch_idxs


class RingBuffer:
    def __init__(self, maxlen):
        self.maxlen = maxlen
        self.length = 0
        self.data = []

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        if idx < 0 or idx >= self.length:
            raise KeyError()
        return self.data[idx]

    def __delitem__(self, subscript):
        del self.data[subscript]
        self.length = len(self.data)

    def append(self, v):
        if self.length < self.maxlen:
            # We have space, simply append to data.
            self.data.append(v)
            self.length += 1

        elif self.length == self.maxlen:
            # No space, remove the first item then append
            self.data.pop(0)
            self.data.append(v)
        else:
            # This should never happen.
            raise nncf.BufferFullError()


def zeroed_observation(observation):
    if hasattr(observation, "shape"):
        return np.zeros(observation.shape)
    if hasattr(observation, "__iter__"):
        out = []
        for x in observation:
            out.append(zeroed_observation(x))
        return out
    return 0.0


class Memory:
    def __init__(self, window_length, ignore_episode_boundaries=False):
        self.window_length = window_length
        self.ignore_episode_boundaries = ignore_episode_boundaries

        self.recent_observations = deque(maxlen=window_length)
        self.recent_terminals = deque(maxlen=window_length)

    def sample(self, batch_size, batch_idxs=None):
        raise NotImplementedError()

    def append(self, observation, action, reward, terminal, training=True):
        self.recent_observations.append(observation)
        self.recent_terminals.append(terminal)

    def get_recent_state(self, current_observation):
        # This code is slightly complicated by the fact that subsequent observations might be
        # from different episodes. We ensure that an experience never spans multiple episodes.
        # This is probably not that important in practice but it seems cleaner.
        state = [current_observation]
        idx = len(self.recent_observations) - 1
        for offset in range(0, self.window_length - 1):
            current_idx = idx - offset
            current_terminal = self.recent_terminals[current_idx - 1] if current_idx - 1 >= 0 else False
            if current_idx < 0 or (not self.ignore_episode_boundaries and current_terminal):
                # The previously handled observation was terminal, don't add the current one.
                # Otherwise we would leak into a different episode.
                break
            state.insert(0, self.recent_observations[current_idx])
        while len(state) < self.window_length:
            state.insert(0, zeroed_observation(state[0]))
        return state

    def get_config(self):
        config = {
            "window_length": self.window_length,
            "ignore_episode_boundaries": self.ignore_episode_boundaries,
        }
        return config


class SequentialMemory(Memory):
    def __init__(self, limit, **kwargs):
        super().__init__(**kwargs)

        self.limit = limit

        # Do not use deque to implement the memory. This data structure may seem convenient but
        # it is way too slow on random access. Instead, we use our own ring buffer implementation.
        self.actions = RingBuffer(limit)
        self.rewards = RingBuffer(limit)
        self.terminals = RingBuffer(limit)
        self.observations = RingBuffer(limit)

    def sample(self, batch_size, batch_idxs=None):
        if batch_idxs is None:
            # Draw random indexes such that we have at least a single entry before each
            # index.
            batch_idxs = sample_batch_indexes(0, self.nb_entries - 1, size=batch_size)
        batch_idxs = np.array(batch_idxs) + 1
        assert np.min(batch_idxs) >= 1
        assert np.max(batch_idxs) < self.nb_entries
        assert len(batch_idxs) == batch_size

        # Create experiences
        experiences = []
        for idx in batch_idxs:
            terminal0 = self.terminals[idx - 2] if idx >= 2 else False
            while terminal0:
                # Skip this transition because the environment was reset here. Select a new, random
                # transition and use this instead. This may cause the batch to contain the same
                # transition twice.
                idx = sample_batch_indexes(1, self.nb_entries, size=1)[0]
                terminal0 = self.terminals[idx - 2] if idx >= 2 else False
            assert 1 <= idx < self.nb_entries

            # This code is slightly complicated by the fact that subsequent observations might be
            # from different episodes. We ensure that an experience never spans multiple episodes.
            # This is probably not that important in practice but it seems cleaner.
            state0 = [self.observations[idx - 1]]
            for offset in range(0, self.window_length - 1):
                current_idx = idx - 2 - offset
                current_terminal = self.terminals[current_idx - 1] if current_idx - 1 > 0 else False
                if current_idx < 0 or (not self.ignore_episode_boundaries and current_terminal):
                    # The previously handled observation was terminal, don't add the current one.
                    # Otherwise we would leak into a different episode.
                    break
                state0.insert(0, self.observations[current_idx])
            while len(state0) < self.window_length:
                state0.insert(0, zeroed_observation(state0[0]))
            action = self.actions[idx - 1]
            reward = self.rewards[idx - 1]
            terminal1 = self.terminals[idx - 1]

            # Okay, now we need to create the follow-up state. This is state0 shifted on timestep
            # to the right. Again, we need to be careful to not include an observation from the next
            # episode if the last state is terminal.
            state1 = [np.copy(x) for x in state0[1:]]
            state1.append(self.observations[idx])

            assert len(state0) == self.window_length
            assert len(state1) == len(state0)
            experiences.append(
                Experience(state0=state0, action=action, reward=reward, state1=state1, terminal1=terminal1)
            )
        assert len(experiences) == batch_size
        return experiences

    def sample_and_split(self, batch_size, batch_idxs=None):
        experiences = self.sample(batch_size, batch_idxs)

        state0_batch = []
        reward_batch = []
        action_batch = []
        terminal1_batch = []
        state1_batch = []
        for e in experiences:
            state0_batch.append(e.state0)
            state1_batch.append(e.state1)
            reward_batch.append(e.reward)
            action_batch.append(e.action)
            terminal1_batch.append(0.0 if e.terminal1 else 1.0)

        # Prepare and validate parameters.
        state0_batch = np.array(state0_batch, "double").reshape(batch_size, -1)
        state1_batch = np.array(state1_batch, "double").reshape(batch_size, -1)
        terminal1_batch = np.array(terminal1_batch, "double").reshape(batch_size, -1)
        reward_batch = np.array(reward_batch, "double").reshape(batch_size, -1)
        action_batch = np.array(action_batch, "double").reshape(batch_size, -1)

        return state0_batch, action_batch, reward_batch, state1_batch, terminal1_batch

    def append(self, observation, action, reward, terminal, training=True):
        super().append(observation, action, reward, terminal, training=training)

        # This needs to be understood as follows: in `observation`, take `action`, obtain `reward`
        # and weather the next state is `terminal` or not.
        if training:
            self.observations.append(observation)
            self.actions.append(action)
            self.rewards.append(reward)
            self.terminals.append(terminal)

    def discard(self, subscript):
        if not isinstance(subscript, slice):
            raise ValueError("discard function only supports input of type slice")

        del self.observations[subscript]
        del self.actions[subscript]
        del self.rewards[subscript]
        del self.terminals[subscript]

    @property
    def nb_entries(self):
        return len(self.observations)

    def get_config(self):
        config = super().get_config()
        config["limit"] = self.limit
        return config



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from contextlib import contextmanager
from pathlib import Path

import psutil

import nncf


def fail_if_symlink(file: Path):
    if file.is_symlink():
        raise nncf.ValidationError("File {} is a symbolic link, aborting.".format(str(file)))


@contextmanager
def safe_open(file: Path, *args, **kwargs):
    """
    Safe function to open file and return a stream.

    For security reasons, should not follow symlinks. Use .resolve() on any Path
    objects before passing them here.

    :param file: The path to the file.
    :return: A file object.
    """
    fail_if_symlink(file)
    with open(str(file), *args, **kwargs) as f:
        yield f


def is_windows():
    return "win32" in sys.platform


def is_linux():
    return "linux" in sys.platform


def get_available_cpu_count(logical: bool = True) -> int:
    """
    Return the number of CPUs in the system.

    :param logical: If False return the number of physical cores only (e.g. hyper thread CPUs are excluded),
      otherwise number of logical cores. Defaults, True.
    :return: Number of CPU.
    """
    try:
        num_cpu = psutil.cpu_count(logical=logical)
        return num_cpu if num_cpu is not None else 1
    except Exception:
        return 1


def get_available_memory_amount() -> int:
    """
    :return: Available memory amount (bytes)
    """
    try:
        return psutil.virtual_memory()[1]
    except Exception:
        return 0



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Type

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFGraphEdge
from nncf.common.graph import NNCFNode
from nncf.common.graph.layer_attributes import GroupNormLayerAttributes
from nncf.common.pruning.symbolic_mask import SymbolicMask
from nncf.common.pruning.tensor_processor import NNCFPruningBaseTensorProcessor
from nncf.common.pruning.utils import get_input_masks
from nncf.common.pruning.utils import identity_mask_propagation
from nncf.common.pruning.utils import is_grouped_conv
from nncf.common.pruning.utils import is_prunable_depthwise_conv
from nncf.common.tensor import NNCFTensor


class BasePruningOp:
    """
    Determines meta operations which aggregate operations having common
    properties of interaction with pruning masks
    """

    subtypes = []
    additional_types = []

    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        """
        :return: accept_pruned_input - can this operation work with pruned input or not
        """
        raise NotImplementedError

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        """
        Propagates the pruning mask through a node using pruning masks of all inputs and the current node (if any).

        :param node: The graph node to propagate mask through it
        :param graph: The model graph to prune
        :param tensor_processor: Interface with tensor processing methods
        """
        raise NotImplementedError

    @classmethod
    def get_all_op_aliases(cls) -> List[str]:
        """
        :return: list of all aliases of types in metatype
        """
        op_types = []
        for subtype in cls.subtypes:
            op_types.extend(subtype.get_all_aliases())
        op_types = list(set(op_types)) + cls.additional_types
        return op_types


class InputPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        return False

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        node.attributes["output_mask"] = None


class OutputPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        return True

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        node.attributes["output_mask"] = None


class IdentityMaskForwardPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        return True

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        identity_mask_propagation(node, graph)


class ConvolutionPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        if is_grouped_conv(node) and not is_prunable_depthwise_conv(node):
            return False
        return True

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_masks = get_input_masks(node, graph)
        output_mask = node.attributes.get("output_mask", None)

        if is_grouped_conv(node):
            output_mask = None
            if is_prunable_depthwise_conv(node):
                output_mask = input_masks[0]

        node.attributes["output_mask"] = output_mask


class TransposeConvolutionPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        if is_grouped_conv(node) and not is_prunable_depthwise_conv(node):
            return False
        return True

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_masks = get_input_masks(node, graph)
        output_mask = node.attributes.get("output_mask", None)

        # In case of group convs we can't prune by output filters
        if is_grouped_conv(node):
            output_mask = None
            if is_prunable_depthwise_conv(node):
                output_mask = input_masks[0]

        node.attributes["output_mask"] = output_mask


class LinearPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        return True

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        output_mask = node.attributes.get("output_mask", None)
        node.attributes["output_mask"] = output_mask


class BatchNormPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        return True

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        identity_mask_propagation(node, graph)


class GroupNormPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        # For Instance Normalization
        return (
            isinstance(node.layer_attributes, GroupNormLayerAttributes)
            and node.layer_attributes.num_groups == node.layer_attributes.num_channels
        )

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        if cls.accept_pruned_input(node):
            identity_mask_propagation(node, graph)
        else:
            node.attributes["output_mask"] = None


class LayerNormPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        return True

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        identity_mask_propagation(node, graph)


class ConcatPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode):
        return True

    @classmethod
    def generate_output_mask(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> Optional[NNCFTensor]:
        """
        Generate output mask from input masks with all None replaced by identity masks.
        If all input masks is None return None.

        :param node: Node to determine it's sources.
        :param graph: NNCF graph to work with.
        :param tensor_processor: Interface with tensor processing methods.
        :return: Filled input masks.
        """
        input_edges = graph.get_input_edges(node)
        previous_nodes = [edge.from_node for edge in input_edges]
        input_masks = [input_node.attributes["output_mask"] for input_node in previous_nodes]
        input_masks = [
            input_mask[node.node_name] if isinstance(input_mask, dict) else input_mask for input_mask in input_masks
        ]

        not_empty_masks: List[NNCFTensor] = [mask for mask in input_masks if mask is not None]
        if not not_empty_masks:
            return None

        first_non_empty_mask = not_empty_masks[0]

        device = first_non_empty_mask.device
        filled_input_masks = []
        for i, mask in enumerate(input_masks):
            if mask is None:
                concat_axis = node.layer_attributes.axis
                concat_dim = input_edges[i].tensor_shape[concat_axis]
                mask = tensor_processor.ones(concat_dim, device)
            filled_input_masks.append(mask)
        result_mask = tensor_processor.concatenate(filled_input_masks, 0)
        return result_mask

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        result_mask = cls.generate_output_mask(node, graph, tensor_processor)
        node.attributes["output_mask"] = result_mask


class SplitPruningOp(BasePruningOp):
    @classmethod
    def match_multiple_output_masks(
        cls, output_masks: List[SymbolicMask], output_edges: List[NNCFGraphEdge], chunk_axis: int
    ) -> Dict["str", SymbolicMask]:
        """
        Match multiple input mask to each next nodes.

        :param output_masks: Given output masks.
        :param output_edges: Given output edges of the node.
        :param chunk_axis: Given the axis on which operation was performed.
        :return: Matched output mask for each next node.
        """
        result_masks = {}
        tmp_output_masks = output_masks.copy()
        tmp_output_masks_shape = [mask.shape[0] for mask in tmp_output_masks]
        for edge in output_edges:
            node_name = edge.to_node.node_name
            idx = tmp_output_masks_shape.index(edge.tensor_shape[chunk_axis])
            result_masks[node_name] = tmp_output_masks.pop(idx)
            tmp_output_masks_shape.pop(idx)

        return result_masks

    @classmethod
    def accept_pruned_input(cls, node: NNCFNode):
        if node.layer_attributes is not None:
            return True
        return False

    @classmethod
    def generate_output_masks(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> Optional[NNCFTensor]:
        """
        Generate output mask from input masks for split/chunk operations.
        If input mask is None return None

        :param node: Node to determine its sources.
        :param graph: NNCF graph to work with.
        :param tensor_processor: Interface with tensor processing methods.
        :return: Filled input masks.
        """
        input_masks = get_input_masks(node, graph)
        if not input_masks:
            return None

        assert len(input_masks) == 1
        input_mask = input_masks[0]

        if not input_mask:
            return None

        chunk_axis = node.layer_attributes.axis

        output_edges = graph.get_output_edges(node)
        output_shapes = [edge.tensor_shape[chunk_axis] for edge in output_edges]

        # if identity split detected
        if len(output_shapes) == 1:
            # propagate as is
            return input_mask

        if input_mask.shape[0] != sum(output_shapes):
            return None

        split_masks = tensor_processor.split(input_mask, output_shapes)
        result_masks = cls.match_multiple_output_masks(split_masks, output_edges, chunk_axis)

        return result_masks

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        result_masks = cls.generate_output_masks(node, graph, tensor_processor)
        node.attributes["output_mask"] = result_masks


class PadPruningOp(IdentityMaskForwardPruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        mode, value = node.layer_attributes.mode, node.layer_attributes.value
        if mode == "constant" and value != 0:
            return False
        return True


class ElementwisePruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        return True

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        input_masks = get_input_masks(node, graph)
        output_mask = input_masks[0]
        if output_mask is not None:
            output_mask = tensor_processor.elementwise_mask_propagation(input_masks)

        node.attributes["output_mask"] = output_mask


class ReshapePruningOp(BasePruningOp):
    @staticmethod
    def _is_flatten(node: NNCFNode) -> bool:
        return len(node.layer_attributes.output_shape) == 2

    @staticmethod
    def _is_not_mixing_dim(node: NNCFNode) -> bool:
        input_shape = node.layer_attributes.input_shape
        output_shape = node.layer_attributes.output_shape

        # TODO(dlyakhov): Cover all corner cases that appear here (ticket 90976)
        if len(input_shape) == len(output_shape) and set(input_shape) == set(output_shape):
            return input_shape == output_shape
        return True

    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        return node.layer_attributes is not None

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        if cls.accept_pruned_input(node):
            if cls._is_flatten(node):
                FlattenPruningOp.mask_propagation(node, graph, tensor_processor)
            elif cls._is_not_mixing_dim(node):
                identity_mask_propagation(node, graph)
            else:
                node.attributes["output_mask"] = None
        else:
            node.attributes["output_mask"] = None


class FlattenPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        if node.layer_attributes is not None:
            return True
        return False

    @classmethod
    def mask_propagation(cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]):
        output_mask = None
        input_masks = get_input_masks(node, graph)
        assert len(input_masks) == 1
        input_mask = input_masks[0]
        if input_mask is not None and node.layer_attributes is not None:
            # We assume all layers known by the mask propagation algo except
            # StopMaskForwardOp have input/output batch dim == 0.
            # Besides, since input_mask is not None thus no StopMaskForwardOp operations
            # was in the path from mask producer node to this node. As all
            # known nodes have input/output batch dim == 0 previous has too.
            flatten_channels = node.layer_attributes.output_shape[1]
            mask_len = input_mask.shape[0]
            assert flatten_channels % mask_len == 0
            output_mask = tensor_processor.repeat(input_mask, repeats=flatten_channels // mask_len)

        node.attributes["output_mask"] = output_mask


class StopMaskForwardPruningOp(BasePruningOp):
    @classmethod
    def accept_pruned_input(cls, node: NNCFNode) -> bool:
        return False

    @classmethod
    def mask_propagation(
        cls, node: NNCFNode, graph: NNCFGraph, tensor_processor: Type[NNCFPruningBaseTensorProcessor]
    ) -> None:
        node.attributes["output_mask"] = None



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, List, Optional, Type, TypeVar

from nncf.common.graph.definitions import NNCFGraphNodeType
from nncf.common.graph.layer_attributes import BaseLayerAttributes
from nncf.common.graph.layer_attributes import ConvolutionLayerAttributes
from nncf.common.graph.operator_metatypes import CONST_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import INPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OUTPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.operator_metatypes import OperatorMetatypeRegistry
from nncf.common.hardware.opset import HWConfigOpName
from nncf.torch.dynamic_graph.graph import DynamicGraph
from nncf.torch.dynamic_graph.structs import NamespaceTarget

ModuleAttributes = TypeVar("ModuleAttributes", bound=BaseLayerAttributes)

PT_OPERATOR_METATYPES = OperatorMetatypeRegistry("operator_metatypes")


class PTOperatorMetatype(OperatorMetatype):
    """
    Base class for grouping PyTorch operators based on their semantic meaning.
    Each derived class represents a single semantic group - for example, AddMetatype would
    group together '__iadd__', '__add__' and '__radd__' operations which all define nodewise
    tensor addition.
    Derived classes also specify which PyTorch functions in which modules should be patched
    so that the entire group of operations is visible in the internal graph.
    Grouping also allows efficient application of HW specifics to compression of
    certain operation groups.

    :param external_op_names: Names of functions registered as operators via @register_operator to be associated
    with this metatype.
    :param module_to_function_names: Names of functions from 'torch.nn.function', 'torch.tensor' and 'torch' modules
    respectively, which are associated with this metatype.
    :param subtypes: List of subtypes of PyTorch operator.
    """

    external_op_names: List[str] = []
    num_expected_input_edges: Optional[int] = None

    module_to_function_names: Dict[NamespaceTarget, List[str]] = {
        NamespaceTarget.TORCH_NN_FUNCTIONAL: [],
        NamespaceTarget.TORCH_TENSOR: [],
        NamespaceTarget.TORCH: [],
    }

    subtypes: List[Type["PTOperatorMetatype"]] = []

    @classmethod
    def get_subtypes(cls) -> List[Type["PTOperatorMetatype"]]:
        return cls.subtypes.copy()

    @classmethod
    def get_all_namespace_to_function_names(cls) -> Dict[NamespaceTarget, List[str]]:
        output = dict(cls.module_to_function_names)
        output[NamespaceTarget.EXTERNAL] = cls.external_op_names
        return output

    @classmethod
    def get_all_aliases(cls) -> List[str]:
        output = set()
        for _, function_names in cls.module_to_function_names.items():
            output = output.union(function_names)
        if cls.external_op_names is not None:
            output = output.union(cls.external_op_names)
        return list(output)

    @classmethod
    def determine_subtype(
        cls, layer_attributes: Optional[BaseLayerAttributes] = None, function_args=None, functions_kwargs=None
    ) -> Optional["PTOperatorSubtype"]:
        matches = []
        for subtype in cls.get_subtypes():
            if subtype.matches(layer_attributes, function_args, functions_kwargs):
                matches.append(subtype)
        assert len(matches) <= 1, "Multiple subtypes match operator call - cannot determine single subtype."
        if not matches:
            return None

        subtype = matches[0]
        nested_subtype = subtype.determine_subtype(layer_attributes, function_args, functions_kwargs)
        if nested_subtype:
            return nested_subtype
        return subtype


class PTOperatorSubtype(PTOperatorMetatype):
    """
    Exact specialization of PTOperatorMetatype that can only be determined via operator argument
    inspection or owning module attribute inspection, and that may have specialized compression method
    configuration other than the one used for general operations having the type of PTOperatorMetatype.
    """

    @classmethod
    def matches(
        cls, layer_attributes: Optional[BaseLayerAttributes] = None, function_args=None, functions_kwargs=None
    ) -> bool:
        raise NotImplementedError


def _is_called_inside_nncf_module(functions_kwargs):
    key = DynamicGraph.IS_CALLED_INSIDE_NNCF_MODULE
    if functions_kwargs is None or key not in functions_kwargs:
        return False
    return functions_kwargs[key]


class PTModuleOperatorSubtype(PTOperatorSubtype):
    @classmethod
    def matches(
        cls, layer_attributes: Optional[BaseLayerAttributes] = None, function_args=None, functions_kwargs=None
    ) -> bool:
        return _is_called_inside_nncf_module(functions_kwargs)


class PTModuleDepthwiseConvOperatorSubtype(PTOperatorSubtype):
    @classmethod
    def matches(
        cls, layer_attributes: Optional[BaseLayerAttributes] = None, function_args=None, functions_kwargs=None
    ) -> bool:
        if not _is_called_inside_nncf_module(functions_kwargs):
            return False
        if not isinstance(layer_attributes, ConvolutionLayerAttributes):
            return False
        if layer_attributes.groups == layer_attributes.in_channels and layer_attributes.in_channels > 1:
            return True
        return False


class PTDepthwiseConvOperatorSubtype(PTOperatorSubtype):
    @classmethod
    def matches(
        cls, layer_attributes: Optional[BaseLayerAttributes] = None, function_args=None, functions_kwargs=None
    ) -> bool:
        if _is_called_inside_nncf_module(functions_kwargs):
            return False
        if not isinstance(layer_attributes, ConvolutionLayerAttributes):
            return False
        if layer_attributes.groups == layer_attributes.in_channels and layer_attributes.in_channels > 1:
            return True
        return False


@PT_OPERATOR_METATYPES.register()
@INPUT_NOOP_METATYPES.register()
class PTInputNoopMetatype(PTOperatorMetatype):
    name = "input_noop"
    external_op_names = [name, NNCFGraphNodeType.INPUT_NODE]


@PT_OPERATOR_METATYPES.register()
@OUTPUT_NOOP_METATYPES.register()
class PTOutputNoopMetatype(PTOperatorMetatype):
    name = "output_noop"
    external_op_names = [name, NNCFGraphNodeType.OUTPUT_NODE]


@PT_OPERATOR_METATYPES.register()
@CONST_NOOP_METATYPES.register()
class PTConstNoopMetatype(PTOperatorMetatype):
    name = "const_noop"
    external_op_names = [name, NNCFGraphNodeType.CONST_NODE]


@PT_OPERATOR_METATYPES.register()
@NOOP_METATYPES.register()
class PTNoopMetatype(PTOperatorMetatype):
    name = "noop"
    external_op_names = [name]
    module_to_function_names = {
        NamespaceTarget.TORCH_NN_FUNCTIONAL: [],
        NamespaceTarget.TORCH_TENSOR: ["contiguous", "clone", "detach", "detach_", "to"],
        NamespaceTarget.TORCH: ["clone", "detach", "detach_"],
    }


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleDepthwiseConv1dSubtype(PTModuleDepthwiseConvOperatorSubtype):
    name = "Conv1DOp"
    hw_config_name = [HWConfigOpName.DEPTHWISECONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv1d"]}
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleConv1dMetatype(PTModuleOperatorSubtype):
    name = "Conv1DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv1d"]}
    subtypes = [PTModuleDepthwiseConv1dSubtype]
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTDepthwiseConv1dSubtype(PTDepthwiseConvOperatorSubtype):
    name = "Conv1DOp"
    hw_config_name = [HWConfigOpName.DEPTHWISECONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv1d"]}
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register()
class PTConv1dMetatype(PTOperatorMetatype):
    name = "Conv1DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv1d"]}
    subtypes = [PTModuleConv1dMetatype, PTDepthwiseConv1dSubtype]
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleDepthwiseConv2dSubtype(PTModuleDepthwiseConvOperatorSubtype):
    name = "Conv2DOp"
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv2d"]}
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleConv2dMetatype(PTModuleOperatorSubtype):
    name = "Conv2DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv2d"]}
    subtypes = [PTModuleDepthwiseConv2dSubtype]
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTDepthwiseConv2dSubtype(PTDepthwiseConvOperatorSubtype):
    name = "Conv2DOp"
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv2d"]}
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register()
class PTConv2dMetatype(PTOperatorMetatype):
    name = "Conv2DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv2d"]}
    subtypes = [PTModuleConv2dMetatype, PTDepthwiseConv2dSubtype]
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleDepthwiseConv3dSubtype(PTModuleDepthwiseConvOperatorSubtype):
    name = "Conv3DOp"
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv3d"]}
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleConv3dMetatype(PTModuleOperatorSubtype):
    name = "Conv3DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv3d"]}
    subtypes = [PTModuleDepthwiseConv3dSubtype]
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTDepthwiseConv3dSubtype(PTDepthwiseConvOperatorSubtype):
    name = "Conv3DOp"
    hw_config_names = [HWConfigOpName.DEPTHWISECONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv3d"]}
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register()
class PTConv3dMetatype(PTOperatorMetatype):
    name = "Conv3DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv3d"]}
    subtypes = [PTModuleConv3dMetatype, PTDepthwiseConv3dSubtype]
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleConvTranspose1dMetatype(PTModuleOperatorSubtype):
    name = "ConvTranspose1DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv_transpose1d"]}
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register()
class PTConvTranspose1dMetatype(PTOperatorMetatype):
    name = "ConvTranspose1DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv_transpose1d"]}
    subtypes = [PTModuleConvTranspose1dMetatype]
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleConvTranspose2dMetatype(PTModuleOperatorSubtype):
    name = "ConvTranspose2DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv_transpose2d"]}
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register()
class PTConvTranspose2dMetatype(PTOperatorMetatype):
    name = "ConvTranspose2DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv_transpose2d"]}
    subtypes = [PTModuleConvTranspose2dMetatype]
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleConvTranspose3dMetatype(PTModuleOperatorSubtype):
    name = "ConvTranspose3DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv_transpose3d"]}
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register()
class PTConvTranspose3dMetatype(PTOperatorMetatype):
    name = "ConvTranspose3DOp"
    hw_config_names = [HWConfigOpName.CONVOLUTION]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["conv_transpose3d"]}
    subtypes = [PTModuleConvTranspose3dMetatype]
    output_channel_axis = 1
    num_expected_input_edges = 2
    weight_port_ids = [1]
    bias_port_id = 2


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleDeformConv2dMetatype(PTModuleOperatorSubtype):
    name = "DeformConv2dOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["deform_conv2d"]}
    num_expected_input_edges = 2
    weight_port_ids = [2]


@PT_OPERATOR_METATYPES.register()
class PTDeformConv2dMetatype(PTOperatorMetatype):
    name = "DeformConv2dOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["deform_conv2d"]}
    subtypes = [PTModuleDeformConv2dMetatype]
    num_expected_input_edges = 4
    weight_port_ids = [2]


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleLinearMetatype(PTModuleOperatorSubtype):
    name = "LinearOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["linear"]}
    hw_config_names = [HWConfigOpName.MATMUL]
    output_channel_axis = -1
    num_expected_input_edges = 2
    weight_port_ids = [1]


@PT_OPERATOR_METATYPES.register()
class PTLinearMetatype(PTOperatorMetatype):
    name = "LinearOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["linear"]}
    hw_config_names = [HWConfigOpName.MATMUL]
    subtypes = [PTModuleLinearMetatype]
    output_channel_axis = -1
    num_expected_input_edges = 2
    weight_port_ids = [1]


@PT_OPERATOR_METATYPES.register()
class PTHardTanhMetatype(PTOperatorMetatype):
    name = "HardTanhOP"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["hardtanh"]}
    num_expected_input_edges = 1


@PT_OPERATOR_METATYPES.register()
class PTHardSwishMetatype(PTOperatorMetatype):
    name = "HardSwishOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["hardswish"]}
    num_expected_input_edges = 1


@PT_OPERATOR_METATYPES.register()
class PTHardSigmoidMetatype(PTOperatorMetatype):
    name = "HardSigmoidOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["hardsigmoid"]}
    num_expected_input_edges = 1


@PT_OPERATOR_METATYPES.register()
class PTTanhMetatype(PTOperatorMetatype):
    name = "TanhOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["tanh"], NamespaceTarget.TORCH: ["tanh"]}
    num_expected_input_edges = 1


@PT_OPERATOR_METATYPES.register()
class PTELUMetatype(PTOperatorMetatype):
    name = "EluOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["elu", "elu_"]}
    num_expected_input_edges = 1


@PT_OPERATOR_METATYPES.register()
class PTPRELUMetatype(PTOperatorMetatype):
    name = "PReluOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["prelu"]}
    num_expected_input_edges = 1


@PT_OPERATOR_METATYPES.register()
class PTLeakyRELUMetatype(PTOperatorMetatype):
    name = "LeakyReluOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["leaky_relu"]}
    num_expected_input_edges = 1


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleLayerNormMetatype(PTModuleOperatorSubtype):
    name = "LayerNormOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["layer_norm"]}
    hw_config_names = [HWConfigOpName.MVN]
    num_expected_input_edges = 1
    weight_port_ids = [2]


@PT_OPERATOR_METATYPES.register()
class PTLayerNormMetatype(PTOperatorMetatype):
    name = "LayerNormOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["layer_norm"]}
    hw_config_names = [HWConfigOpName.MVN]
    subtypes = [PTModuleLayerNormMetatype]
    num_expected_input_edges = 1
    weight_port_ids = [2]


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleGroupNormMetatype(PTModuleOperatorSubtype):
    name = "GroupNormOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["group_norm"]}
    hw_config_names = [HWConfigOpName.MVN]
    weight_port_ids = [2]


@PT_OPERATOR_METATYPES.register()
class PTGroupNormMetatype(PTOperatorMetatype):
    name = "GroupNormOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["group_norm"]}
    hw_config_names = [HWConfigOpName.MVN]
    subtypes = [PTModuleGroupNormMetatype]
    weight_port_ids = [2]


@PT_OPERATOR_METATYPES.register()
class PTGELUMetatype(PTOperatorMetatype):
    name = "GeluOp"
    hw_config_names = [HWConfigOpName.GELU]
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["gelu"]}


@PT_OPERATOR_METATYPES.register()
class PTSILUMetatype(PTOperatorMetatype):
    name = "SiluOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["silu"]}


@PT_OPERATOR_METATYPES.register()
class PTSigmoidMetatype(PTOperatorMetatype):
    name = "SigmoidOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_NN_FUNCTIONAL: ["sigmoid"],
        NamespaceTarget.TORCH_TENSOR: ["sigmoid"],
        NamespaceTarget.TORCH: ["sigmoid"],
    }


@PT_OPERATOR_METATYPES.register()
class PTAddMetatype(PTOperatorMetatype):
    name = "AddOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["add", "__add__", "__iadd__", "__radd__"],
        NamespaceTarget.TORCH: ["add"],
    }
    hw_config_names = [HWConfigOpName.ADD]
    num_expected_input_edges = 2


@PT_OPERATOR_METATYPES.register()
class PTSubMetatype(PTOperatorMetatype):
    name = "SubOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["sub", "__sub__", "__isub__", "__rsub__"],
        NamespaceTarget.TORCH: ["sub"],
    }
    hw_config_names = [HWConfigOpName.SUBTRACT]
    num_expected_input_edges = 2


@PT_OPERATOR_METATYPES.register()
class PTMulMetatype(PTOperatorMetatype):
    name = "MulOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["mul", "__mul__", "__imul__", "__rmul__"],
        NamespaceTarget.TORCH: ["mul"],
    }
    hw_config_names = [HWConfigOpName.MULTIPLY]
    num_expected_input_edges = 2


@PT_OPERATOR_METATYPES.register()
class PTDivMetatype(PTOperatorMetatype):
    name = "DivOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: [
            "div",
            "__div__",
            "__idiv__",
            "__rdiv__",
            "__truediv__",
            "__itruediv__",
            "__rtruediv__",
        ],
        NamespaceTarget.TORCH: ["div"],
    }
    hw_config_names = [HWConfigOpName.DIVIDE]
    num_expected_input_edges = 2


@PT_OPERATOR_METATYPES.register()
class PTFloorDivMetatype(PTOperatorMetatype):
    name = "FloordivOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["__floordiv__", "__ifloordiv__", "__rfloordiv__"],
        NamespaceTarget.TORCH: ["floor_divide"],
    }
    num_expected_input_edges = 2


@PT_OPERATOR_METATYPES.register()
class PTExpMetatype(PTOperatorMetatype):
    name = "ExpOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["exp"],
        NamespaceTarget.TORCH: ["exp"],
    }


@PT_OPERATOR_METATYPES.register()
class PTLogMetatype(PTOperatorMetatype):
    name = "LogOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["log"],
        NamespaceTarget.TORCH: ["log"],
    }


@PT_OPERATOR_METATYPES.register()
class PTAbsMetatype(PTOperatorMetatype):
    name = "AbsOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["abs", "__abs__"],
        NamespaceTarget.TORCH: ["abs"],
    }


@PT_OPERATOR_METATYPES.register()
class PTErfMetatype(PTOperatorMetatype):
    name = "ErfOp"
    module_to_function_names = {
        NamespaceTarget.TORCH: ["erf"],
    }


@PT_OPERATOR_METATYPES.register()
class PTMatMulMetatype(PTOperatorMetatype):
    name = "MatMulOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["matmul", "__matmul__", "__rmatmul__"],
        NamespaceTarget.TORCH: ["matmul", "bmm", "mm"],
    }
    hw_config_names = [HWConfigOpName.MATMUL]
    num_expected_input_edges = 2
    weight_port_ids = [0, 1]


@PT_OPERATOR_METATYPES.register()
class PTAddmmMetatype(PTOperatorMetatype):
    name = "MatMulOp"
    module_to_function_names = {NamespaceTarget.TORCH: ["addmm", "baddbmm"]}
    hw_config_names = [HWConfigOpName.MATMUL]
    # 0-th arg to the baddbmm is basically a (b)ias to be (add)ed to the (bmm) operation,
    # presuming that most runtime implementations will fuse the bias addition into the matrix multiplication
    # and therefore won't quantize the bias input, as this would break the hardware-fused pattern.
    ignored_input_ports: List[int] = [0]
    num_expected_input_edges = 2
    weight_port_ids = [1, 2]


@PT_OPERATOR_METATYPES.register()
class PTMeanMetatype(PTOperatorMetatype):
    name = "MeanOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["mean"]}
    hw_config_names = [HWConfigOpName.REDUCEMEAN]


@PT_OPERATOR_METATYPES.register()
class PTRoundMetatype(PTOperatorMetatype):
    name = "RoundOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["round"]}


@PT_OPERATOR_METATYPES.register()
class PTDropoutMetatype(PTOperatorMetatype):
    name = "DropoutOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["dropout"]}


@PT_OPERATOR_METATYPES.register()
class PTThresholdMetatype(PTOperatorMetatype):
    name = "ThresholdOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["threshold"]}


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleBatchNormMetatype(PTModuleOperatorSubtype):
    name = "BatchNormOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["batch_norm"]}


@PT_OPERATOR_METATYPES.register()
class PTBatchNormMetatype(PTOperatorMetatype):
    name = "BatchNormOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["batch_norm"]}
    subtypes = [PTModuleBatchNormMetatype]
    weight_port_ids = [3]
    bias_port_id = 4


@PT_OPERATOR_METATYPES.register()
class PTAvgPool2dMetatype(PTOperatorMetatype):
    name = "AvgPool2DOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["avg_pool2d", "adaptive_avg_pool2d"]}
    hw_config_names = [HWConfigOpName.AVGPOOL]


@PT_OPERATOR_METATYPES.register()
class PTAvgPool3dMetatype(PTOperatorMetatype):
    name = "AvgPool3DOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["avg_pool3d", "adaptive_avg_pool3d"]}
    hw_config_names = [HWConfigOpName.AVGPOOL]


class PTAdaptiveMaxPool1dMetatype(PTOperatorMetatype):
    name = "AdaptiveMaxPool1DOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["adaptive_max_pool1d"]}


@PT_OPERATOR_METATYPES.register()
class PTAdaptiveMaxPool2dMetatype(PTOperatorMetatype):
    name = "AdaptiveMaxPool2DOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["adaptive_max_pool2d"]}


@PT_OPERATOR_METATYPES.register()
class PTAdaptiveMaxPool3dMetatype(PTOperatorMetatype):
    name = "AdaptiveMaxPool3DOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["adaptive_max_pool3d"]}


class PTMaxPool1dMetatype(PTOperatorMetatype):
    name = "MaxPool1DOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["max_pool1d"]}
    hw_config_names = [HWConfigOpName.MAXPOOL]


@PT_OPERATOR_METATYPES.register()
class PTMaxPool2dMetatype(PTOperatorMetatype):
    name = "MaxPool2DOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["max_pool2d"]}
    hw_config_names = [HWConfigOpName.MAXPOOL]


@PT_OPERATOR_METATYPES.register()
class PTMaxPool3dMetatype(PTOperatorMetatype):
    name = "MaxPool3DOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["max_pool3d"]}
    hw_config_names = [HWConfigOpName.MAXPOOL]


@PT_OPERATOR_METATYPES.register()
class PTMaxUnpool1dMetatype(PTOperatorMetatype):
    name = "MaxUnPool1DOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["max_unpool1d"]}


@PT_OPERATOR_METATYPES.register()
class PTMaxUnpool2dMetatype(PTOperatorMetatype):
    name = "MaxUnPool2DOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["max_unpool2d"]}


@PT_OPERATOR_METATYPES.register()
class PTMaxUnpool3dMetatype(PTOperatorMetatype):
    name = "MaxUnPool3DOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["max_unpool3d"]}


@PT_OPERATOR_METATYPES.register()
class PTPadMetatype(PTOperatorMetatype):
    name = "PadOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["pad"]}


@PT_OPERATOR_METATYPES.register()
class PTCatMetatype(PTOperatorMetatype):
    name = "CatOp"
    module_to_function_names = {NamespaceTarget.TORCH: ["cat", "stack"]}
    hw_config_names = [HWConfigOpName.CONCAT]


@PT_OPERATOR_METATYPES.register()
class PTRELUMetatype(PTOperatorMetatype):
    name = "ReluOp"
    module_to_function_names = {NamespaceTarget.TORCH: ["relu", "relu_"]}


@PT_OPERATOR_METATYPES.register()
class PTRELU6Metatype(PTOperatorMetatype):
    name = "Relu6Op"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["relu6"]}


@PT_OPERATOR_METATYPES.register()
class PTMaxMetatype(PTOperatorMetatype):
    name = "MaxOp"
    module_to_function_names = {NamespaceTarget.TORCH: ["max"]}
    hw_config_names = [HWConfigOpName.MAXIMUM, HWConfigOpName.REDUCEMAX]


@PT_OPERATOR_METATYPES.register()
class PTMinMetatype(PTOperatorMetatype):
    name = "MinOp"
    module_to_function_names = {NamespaceTarget.TORCH: ["min"]}
    hw_config_names = [HWConfigOpName.MINIMUM]


@PT_OPERATOR_METATYPES.register()
class PTTransposeMetatype(PTOperatorMetatype):
    name = "TransposeOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["transpose", "permute", "transpose_"],
        NamespaceTarget.TORCH: ["transpose"],
    }
    hw_config_names = [HWConfigOpName.TRANSPOSE]


@PT_OPERATOR_METATYPES.register()
class PTGatherMetatype(PTOperatorMetatype):
    name = "GatherOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["index_select", "__getitem__"],
        NamespaceTarget.TORCH: ["gather", "index_select", "where"],
    }


@PT_OPERATOR_METATYPES.register()
class PTScatterMetatype(PTOperatorMetatype):
    name = "ScatterOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["scatter", "masked_fill", "masked_fill_"]}


@PT_OPERATOR_METATYPES.register()
class PTReshapeMetatype(PTOperatorMetatype):
    name = "ReshapeOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["reshape", "view", "flatten", "unsqueeze"],
        NamespaceTarget.TORCH: ["flatten", "unsqueeze"],
    }
    hw_config_names = [HWConfigOpName.RESHAPE, HWConfigOpName.UNSQUEEZE, HWConfigOpName.FLATTEN]


@PT_OPERATOR_METATYPES.register()
class PTSqueezeMetatype(PTOperatorMetatype):
    name = "SqueezeOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["squeeze"],
        NamespaceTarget.TORCH: ["squeeze"],
    }
    hw_config_names = [HWConfigOpName.SQUEEZE]


@PT_OPERATOR_METATYPES.register()
class PTSplitMetatype(PTOperatorMetatype):
    name = "SplitOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_NN_FUNCTIONAL: [],
        NamespaceTarget.TORCH_TENSOR: ["split", "chunk", "unbind"],
        NamespaceTarget.TORCH: ["split", "chunk", "unbind"],
    }
    hw_config_names = [HWConfigOpName.SPLIT, HWConfigOpName.CHUNK]


@PT_OPERATOR_METATYPES.register()
class PTExpandMetatype(PTOperatorMetatype):
    name = "ExpandOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["expand"]}


@PT_OPERATOR_METATYPES.register()
class PTExpandAsMetatype(PTOperatorMetatype):
    name = "ExpandAsOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["expand_as"]}


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleEmbeddingMetatype(PTModuleOperatorSubtype):
    name = "EmbeddingOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["embedding"]}
    hw_config_names = [HWConfigOpName.EMBEDDING]
    weight_port_ids = [1]


@PT_OPERATOR_METATYPES.register()
class PTEmbeddingMetatype(PTOperatorMetatype):
    name = "EmbeddingOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["embedding"]}
    hw_config_names = [HWConfigOpName.EMBEDDING]
    subtypes = [PTModuleEmbeddingMetatype]
    weight_port_ids = [1]


@PT_OPERATOR_METATYPES.register(is_subtype=True)
class PTModuleEmbeddingBagMetatype(PTModuleOperatorSubtype):
    name = "EmbeddingBagOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["embedding_bag"]}
    hw_config_names = [HWConfigOpName.EMBEDDINGBAG]
    weight_port_ids = [1]


@PT_OPERATOR_METATYPES.register()
class PTEmbeddingBagMetatype(PTOperatorMetatype):
    name = "EmbeddingBagOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["embedding_bag"]}
    hw_config_names = [HWConfigOpName.EMBEDDINGBAG]
    subtypes = [PTModuleEmbeddingBagMetatype]
    weight_port_ids = [1]


@PT_OPERATOR_METATYPES.register()
class PTSoftmaxMetatype(PTOperatorMetatype):
    name = "SoftmaxOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["softmax"]}


@PT_OPERATOR_METATYPES.register()
class PTLessMetatype(PTOperatorMetatype):
    name = "LessOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["__lt__"]}
    hw_config_names = [HWConfigOpName.LESS]


@PT_OPERATOR_METATYPES.register()
class PTLessEqualMetatype(PTOperatorMetatype):
    name = "LessEqualOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["__le__"]}
    hw_config_names = [HWConfigOpName.LESSEQUAL]


@PT_OPERATOR_METATYPES.register()
class PTGreaterMetatype(PTOperatorMetatype):
    name = "GreaterOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["__gt__"]}
    hw_config_names = [HWConfigOpName.GREATER]


@PT_OPERATOR_METATYPES.register()
class PTGreaterEqualMetatype(PTOperatorMetatype):
    name = "GreaterEqualOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["__ge__"]}
    hw_config_names = [HWConfigOpName.GREATEREQUAL]


@PT_OPERATOR_METATYPES.register()
class PTModMetatype(PTOperatorMetatype):
    name = "ModOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["__mod__"]}
    hw_config_names = [HWConfigOpName.FLOORMOD]


@PT_OPERATOR_METATYPES.register()
class PTEqualsMetatype(PTOperatorMetatype):
    name = "EqualsOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["__eq__"]}
    hw_config_names = [HWConfigOpName.EQUAL]


@PT_OPERATOR_METATYPES.register()
class PTNotEqualMetatype(PTOperatorMetatype):
    name = "NotEqualOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["__ne__"]}
    hw_config_names = [HWConfigOpName.NOTEQUAL]


@PT_OPERATOR_METATYPES.register()
class PTLogicalOrMetatype(PTOperatorMetatype):
    name = "LogicalOrOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["__or__", "__ior__", "__ror__"]}
    hw_config_names = [HWConfigOpName.LOGICALOR]


@PT_OPERATOR_METATYPES.register()
class PTLogicalXorMetatype(PTOperatorMetatype):
    name = "LogicalXorOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["__xor__", "__ixor__", "__rxor__"]}
    hw_config_names = [HWConfigOpName.LOGICALXOR]


@PT_OPERATOR_METATYPES.register()
class PTLogicalAndMetatype(PTOperatorMetatype):
    name = "LogicalAndOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["__and__", "__iand__", "__rand__"]}
    hw_config_names = [HWConfigOpName.LOGICALAND]


@PT_OPERATOR_METATYPES.register()
class PTLogicalNotMetatype(PTOperatorMetatype):
    name = "LogicalNotOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["logical_not_", "__invert__"]}
    hw_config_names = [HWConfigOpName.LOGICALNOT]


@PT_OPERATOR_METATYPES.register()
class PTNegativeMetatype(PTOperatorMetatype):
    name = "NegativeOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["neg", "__neg__"],
        NamespaceTarget.TORCH: ["neg"],
    }


@PT_OPERATOR_METATYPES.register()
class PTPowerMetatype(PTOperatorMetatype):
    name = "PowerOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["pow", "__pow__", "__ipow__", "__rpow__"],
        NamespaceTarget.TORCH: ["pow"],
    }
    hw_config_names = [HWConfigOpName.POWER]


@PT_OPERATOR_METATYPES.register()
class PTSqrtMetatype(PTOperatorMetatype):
    name = "SqrtOp"
    module_to_function_names = {
        NamespaceTarget.TORCH_TENSOR: ["sqrt", "sqrt_"],
        NamespaceTarget.TORCH: ["sqrt", "sqrt_"],
    }
    hw_config_names = [HWConfigOpName.POWER]


@PT_OPERATOR_METATYPES.register()
class PTInterpolateMetatype(PTOperatorMetatype):
    name = "InterpolateOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["interpolate"]}
    hw_config_names = [HWConfigOpName.INTERPOLATE]
    num_expected_input_edges = 1


@PT_OPERATOR_METATYPES.register()
class PTRepeatMetatype(PTOperatorMetatype):
    name = "RepeatOp"
    module_to_function_names = {NamespaceTarget.TORCH: ["repeat_interleave"]}
    hw_config_names = [HWConfigOpName.TILE]
    num_expected_input_edges = 1


@PT_OPERATOR_METATYPES.register()
class PTPixelShuffleMetatype(PTOperatorMetatype):
    name = "PixelShuffleOp"
    module_to_function_names = {NamespaceTarget.TORCH_NN_FUNCTIONAL: ["pixel_shuffle"]}
    num_expected_input_edges = 1


@PT_OPERATOR_METATYPES.register()
class PTSumMetatype(PTOperatorMetatype):
    name = "SumOp"
    module_to_function_names = {NamespaceTarget.TORCH_TENSOR: ["sum"], NamespaceTarget.TORCH: ["sum"]}
    hw_config_names = [HWConfigOpName.REDUCESUM]
    num_expected_input_edges = 1


@PT_OPERATOR_METATYPES.register()
class PTReduceL2(PTOperatorMetatype):
    name = "ReduceL2"
    module_to_function_names = {
        NamespaceTarget.TORCH_NN_FUNCTIONAL: ["normalize"],  # note: normalize is for general L_p normalization
    }
    hw_config_names = [HWConfigOpName.REDUCEL2]
    num_expected_input_edges = 1


def get_operator_metatypes() -> List[Type[OperatorMetatype]]:
    """
    Returns a list of the operator metatypes.

    :return: List of operator metatypes .
    """
    return list(PT_OPERATOR_METATYPES.registry_dict.values())


OPERATORS_WITH_WEIGHTS_METATYPES = [
    PTModuleConv1dMetatype,
    PTModuleConv2dMetatype,
    PTModuleConv3dMetatype,
    PTModuleDepthwiseConv1dSubtype,
    PTModuleDepthwiseConv2dSubtype,
    PTModuleDepthwiseConv3dSubtype,
    PTModuleLinearMetatype,
    PTModuleBatchNormMetatype,
    PTModuleGroupNormMetatype,
    PTModuleLayerNormMetatype,
    PTModuleConvTranspose1dMetatype,
    PTModuleConvTranspose2dMetatype,
    PTModuleConvTranspose3dMetatype,
    PTModuleEmbeddingMetatype,
    PTModuleEmbeddingBagMetatype,
]

UNIFICATION_PRODUCING_METATYPES = [
    PTModuleConv1dMetatype,
    PTModuleConv2dMetatype,
    PTModuleConv3dMetatype,
    PTModuleDepthwiseConv1dSubtype,
    PTModuleDepthwiseConv2dSubtype,
    PTModuleDepthwiseConv3dSubtype,
    PTModuleConvTranspose1dMetatype,
    PTModuleConvTranspose2dMetatype,
    PTModuleConvTranspose3dMetatype,
    PTModuleLinearMetatype,
]

OP_NAMES_WITH_WEIGHTS = [x for meta in OPERATORS_WITH_WEIGHTS_METATYPES for x in meta.get_all_aliases()]

QUANTIZE_NODE_TYPES = ["symmetric_quantize", "asymmetric_quantize"]

# These metatypes mix outputs for different samples into one axis.
# If reducers and aggregators collect statistics at the output of the following operations,
# assuming that 0-axis is batch axis, they get only 1 value instead of batch_size values.
# It could lead to inaccurate/incorrect statistics result.
OPERATIONS_OUTPUT_HAS_NO_BATCH_AXIS = [
    PTEmbeddingMetatype,
    PTEmbeddingBagMetatype,
    PTModuleEmbeddingBagMetatype,
    PTModuleEmbeddingMetatype,
]



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator
from typing import Callable, List, TypeVar, Union

import numpy as np

TTensor = TypeVar("TTensor")


def get_subset_indices(errors: List[float], subset_size: int) -> List[int]:
    """
    Returns `subset_size` indices of elements in the `errors` list
    that have the biggest error value. Returned indices are sorted in
    ascending order.

    :param errors: A list of errors.
    :param subset_size: A number of returned indices.
    :return: Indices of elements in the `errors` list which have the biggest error value.
    """
    ordered_indices = [idx for idx, _ in sorted(enumerate(errors), key=operator.itemgetter(1), reverse=True)]
    end_index = min(subset_size, len(ordered_indices))
    return sorted(ordered_indices[:end_index])


def get_subset_indices_pot_version(errors: List[float], subset_size: int) -> List[int]:
    """
    POT implementation of the `get_subset_indices()` method.
    """
    ordered_indices = np.flip(np.argsort(errors)).tolist()
    end_index = min(subset_size, len(ordered_indices))
    return sorted(ordered_indices[:end_index])


def select_subset(
    subset_size: int,
    reference_values_for_each_item: Union[List[float], List[List[TTensor]]],
    approximate_values_for_each_item: Union[List[float], List[List[TTensor]]],
    error_fn: Callable[[Union[float, List[TTensor]], Union[float, List[TTensor]]], float],
) -> List[int]:
    """
    Selects first `subset_size` indices of data items for which `error_fn` function gives maximal value.
    Assumes that `reference_values_for_each_item` and `approximate_values_for_each_item` lists have same
    number of items.

    :param subset_size: Number of indices that will be selected. The `len(reference_values_for_each_item)`
        indices will be selected if `subset_size` parameter is greater than the number of elements in
        the `reference_values_for_each_item`.
    :param reference_values_for_each_item: List of reference values.
    :param approximate_values_for_each_item: List of approximate values.
    :param error_fn: A function used to calculate difference between `reference_values_for_each_item[i]`
        and `approximate_values_for_each_item[i]` list.
    :return: First `subset_size` indices of data items for which `error_fn` function gives maximal value.
    """
    errors = [
        error_fn(ref_val, approx_val)
        for ref_val, approx_val in zip(reference_values_for_each_item, approximate_values_for_each_item)
    ]
    subset_indices = get_subset_indices_pot_version(errors, subset_size)

    return subset_indices



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

NNCF_PACKAGE_ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
HW_CONFIG_RELATIVE_DIR = "common/hardware/configs"

NNCF_CACHE_PATH = Path.home() / Path(".cache/nncf/")
CACHE_MODELS_PATH = NNCF_CACHE_PATH / Path("models")

# Environment variables below, if set, mark the execution environment
# so that certain actions within NNCF proper, such as telemetry event collection or
# debug dumps, are performed or not performed
NNCF_CI_ENV_VAR_NAME = "NNCF_CI"  # Must be set in CI environments
NNCF_DEV_ENV_VAR_NAME = "NNCF_DEV"  # Must be set in environments of the NNCF dev team machines



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict

from nncf.common.graph import NNCFNodeName
from nncf.common.quantization.quantizer_setup import QuantizationPointId
from nncf.common.quantization.quantizer_setup import SingleConfigQuantizerSetup


class CompressionRatioCalculator:
    """
    Calculates compression ratio - ratio between bits complexity of fully INT8 model and mixed-precision lower-bit one.
    Bit complexity of the model is a sum of bit complexities for each quantized layer, which are a multiplication of
    FLOPS for the layer by number of bits for its quantization. The compression ratio can be used for estimation of
    performance boost for quantized model.
    """

    DEFAULT_NUMBER_OF_BITS = 8

    def __init__(
        self,
        flops_per_weighted_module_node: Dict[NNCFNodeName, int],
        quantizer_setup: SingleConfigQuantizerSetup,
        weight_qp_id_per_activation_qp_id: Dict[QuantizationPointId, QuantizationPointId],
    ):
        self._weight_qp_id_per_activation_qp_id = weight_qp_id_per_activation_qp_id
        self._flops_per_weight_qp_id: Dict[QuantizationPointId, float] = {}
        for qp_id, qp in quantizer_setup.quantization_points.items():
            if qp.is_weight_quantization_point():
                target_node_name = qp.insertion_point.target_node_name
                self._flops_per_weight_qp_id[qp_id] = flops_per_weighted_module_node[target_node_name]
        self.maximum_bits_complexity = sum(self._flops_per_weight_qp_id.values()) * self.DEFAULT_NUMBER_OF_BITS

    def run_for_quantizer_setup(self, quantizer_setup: SingleConfigQuantizerSetup) -> float:
        """
        Calculates compression ratio for a given quantizer setup with
        :param: quantizer_setup: setup with information quantization points
        :returns: compression ratio of mixed-precision model by relation to fully INT8
        """
        quantization_points = quantizer_setup.quantization_points
        weight_qps = list(filter(lambda pair: pair[1].is_weight_quantization_point(), quantization_points.items()))
        bits_complexity = 0
        for w_qp_id, w_qp in weight_qps:
            wq_num_bits = w_qp.qconfig.num_bits
            a_qp_id = self._weight_qp_id_per_activation_qp_id[w_qp_id]
            a_qp = quantization_points[a_qp_id]
            aq_num_bits = a_qp.qconfig.num_bits
            num_bits = max(wq_num_bits, aq_num_bits)
            bits_complexity += num_bits * self._flops_per_weight_qp_id[w_qp_id]
        return self.maximum_bits_complexity / bits_complexity



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.openvino.graph.metatypes import openvino_metatypes as ov_metatypes

QUANTIZE_AGNOSTIC_OPERATIONS = [
    ov_metatypes.OVMaxPoolMetatype,
    ov_metatypes.OVReduceMaxMetatype,
    ov_metatypes.OVReshapeMetatype,
    ov_metatypes.OVSqueezeMetatype,
    ov_metatypes.OVUnsqueezeMetatype,
    ov_metatypes.OVSplitMetatype,
    ov_metatypes.OVVariadicSplitMetatype,
    ov_metatypes.OVTransposeMetatype,
    ov_metatypes.OVTileMetatype,
    ov_metatypes.OVStridedSliceMetatype,
    ov_metatypes.OVShuffleChannelsMetatype,
    ov_metatypes.OVBroadcastMetatype,
    ov_metatypes.OVPadMetatype,
    ov_metatypes.OVMinimumMetatype,
    ov_metatypes.OVMaximumMetatype,
    ov_metatypes.OVConvertLikeMetatype,
    ov_metatypes.OVGatherMetatype,
    ov_metatypes.OVGatherNDMetatype,
    ov_metatypes.OVGatherElementsMetatype,
    ov_metatypes.OVScatterUpdateMetatype,
    ov_metatypes.OVScatterNDUpdateMetatype,
    ov_metatypes.OVScatterElementsUpdateMetatype,
    ov_metatypes.OVDepthToSpaceMetatype,
    ov_metatypes.OVSpaceToDepthMetatype,
    ov_metatypes.OVBatchToSpaceMetatype,
    ov_metatypes.OVSpaceToBatchMetatype,
    # OVReluMetatype is not considered to be QUANTIZATION_AGNOSTIC, because:
    # 1. Runtime doesn't provide performance benefits by quantizing the stand-alone RELU's (ticket: 59548)
    # 2. It's frequently better for the end accuracy to have quantizers set up after the RELU
    # so that the input distribution to the quantizer is non-negative
    # and we can therefore have better quantization resolution while preserving the original dynamic range
]


INPUTS_QUANTIZABLE_OPERATIONS = [
    ov_metatypes.OVConvolutionMetatype,
    ov_metatypes.OVGroupConvolutionMetatype,
    ov_metatypes.OVDepthwiseConvolutionMetatype,
    ov_metatypes.OVConvolutionBackpropDataMetatype,
    ov_metatypes.OVGroupConvolutionBackpropDataMetatype,
    ov_metatypes.OVMatMulMetatype,
    ov_metatypes.OVBatchNormMetatype,
    ov_metatypes.OVAddMetatype,
    ov_metatypes.OVSubtractMetatype,
    ov_metatypes.OVMultiplyMetatype,
    ov_metatypes.OVDivideMetatype,
    ov_metatypes.OVMaximumMetatype,
    ov_metatypes.OVMinimumMetatype,
    ov_metatypes.OVAvgPoolMetatype,
    ov_metatypes.OVReduceMeanMetatype,
    ov_metatypes.OVMVNMetatype,
    ov_metatypes.OVNormalizeL2Metatype,
    ov_metatypes.OVInterpolateMetatype,
    ov_metatypes.OVPowerMetatype,
    ov_metatypes.OVFloorModMetatype,
    ov_metatypes.OVLessMetatype,
    ov_metatypes.OVLessEqualMetatype,
    ov_metatypes.OVGreaterMetatype,
    ov_metatypes.OVGreaterEqualMetatype,
    ov_metatypes.OVEqualMetatype,
    ov_metatypes.OVNotEqualMetatype,
    ov_metatypes.OVLogicalNotMetatype,
    ov_metatypes.OVLogicalAndMetatype,
    ov_metatypes.OVLogicalOrMetatype,
    ov_metatypes.OVLogicalXorMetatype,
    ov_metatypes.OVSquaredDifferenceMetatype,
    ov_metatypes.OVLSTMSequenceMetatype,
    ov_metatypes.OVGRUSequenceMetatype,
    ov_metatypes.OVGroupNormalizationMetatype,
    ov_metatypes.OVScaledDotProductAttentionMetatype,
]


FAKE_QUANTIZE_OPERATIONS = [ov_metatypes.OVFakeQuantizeMetatype, ov_metatypes.OVFakeConvertMetatype]


CONSTANT_OPERATIONS = [
    ov_metatypes.OVConstantMetatype,
]


SHAPEOF_OPERATIONS = [
    ov_metatypes.OVShapeOfMetatype,
]


# TODO(andrey-churkin): Can we provide a more suitable name for this variable?
LINEAR_OPERATIONS = [
    ov_metatypes.OVConvolutionMetatype,
    ov_metatypes.OVGroupConvolutionMetatype,
    ov_metatypes.OVDepthwiseConvolutionMetatype,
    ov_metatypes.OVConvolutionBackpropDataMetatype,
    ov_metatypes.OVGroupConvolutionBackpropDataMetatype,
    ov_metatypes.OVDeformableConvolutionMetatype,
    ov_metatypes.OVMatMulMetatype,
]


ATOMIC_ACTIVATIONS_OPERATIONS = [
    ov_metatypes.OVReluMetatype,
    ov_metatypes.OVClampMetatype,
    ov_metatypes.OVEluMetatype,
    ov_metatypes.OVPReluMetatype,
    ov_metatypes.OVSigmoidMetatype,
    ov_metatypes.OVHSigmoidMetatype,
    ov_metatypes.OVHardSigmoidMetatype,
    ov_metatypes.OVSwishMetatype,
    ov_metatypes.OVHSwishMetatype,
]


ARITHMETIC_OPERATIONS = [
    ov_metatypes.OVAddMetatype,
    ov_metatypes.OVSubtractMetatype,
    ov_metatypes.OVMultiplyMetatype,
    ov_metatypes.OVDivideMetatype,
]


ELEMENTWISE_OPERATIONS = [
    ov_metatypes.OVAddMetatype,
    ov_metatypes.OVMultiplyMetatype,
    ov_metatypes.OVSubtractMetatype,
    ov_metatypes.OVDivideMetatype,
    ov_metatypes.OVLessMetatype,
    ov_metatypes.OVLessEqualMetatype,
    ov_metatypes.OVGreaterMetatype,
    ov_metatypes.OVGreaterEqualMetatype,
    ov_metatypes.OVEqualMetatype,
    ov_metatypes.OVNotEqualMetatype,
    ov_metatypes.OVFloorModMetatype,
    ov_metatypes.OVLogicalOrMetatype,
    ov_metatypes.OVLogicalXorMetatype,
    ov_metatypes.OVLogicalAndMetatype,
    ov_metatypes.OVMaximumMetatype,
    ov_metatypes.OVMinimumMetatype,
]


BATCH_NORMALIZATION_OPERATIONS = [
    ov_metatypes.OVBatchNormMetatype,
]


# Keep in mind that having a metatype in this list is necessary for
# the operation to be considered as an "operation with weights", but
# it's not sufficient. For example, the MatMul operation generally has
# weights when it represents a fully connected layer. However, it can
# also be without weights. Therefore, an additional condition is needed
# to check for the existence of weights.
OPERATIONS_WITH_WEIGHTS = [
    ov_metatypes.OVConvolutionMetatype,
    ov_metatypes.OVGroupConvolutionMetatype,
    ov_metatypes.OVDepthwiseConvolutionMetatype,
    ov_metatypes.OVConvolutionBackpropDataMetatype,
    ov_metatypes.OVGroupConvolutionBackpropDataMetatype,
    ov_metatypes.OVMatMulMetatype,
    ov_metatypes.OVLSTMSequenceMetatype,
    ov_metatypes.OVGRUSequenceMetatype,
    ov_metatypes.OVEmbeddingMetatype,
]


# The same comment as mentioned above makes sense.
OPERATIONS_WITH_CONST_PORT_ID = [
    *OPERATIONS_WITH_WEIGHTS,
    ov_metatypes.OVAddMetatype,
]


# Contains the operation metatypes for which bias can be applied.
# Limited operations scope
OPERATIONS_WITH_BIAS_REDUCED = [
    ov_metatypes.OVConvolutionMetatype,
    ov_metatypes.OVMatMulMetatype,
]

OPERATIONS_WITH_BIAS = [*OPERATIONS_WITH_BIAS_REDUCED, ov_metatypes.OVDepthwiseConvolutionMetatype]

CONV_OPERATIONS = [
    ov_metatypes.OVConvolutionMetatype,
    ov_metatypes.OVDepthwiseConvolutionMetatype,
    ov_metatypes.OVGroupConvolutionMetatype,
    ov_metatypes.OVConvolutionBackpropDataMetatype,
    ov_metatypes.OVGroupConvolutionBackpropDataMetatype,
]

# These metatypes mix outputs for different samples into one axis.
# If reducers and aggregators collect statistics at the output of the following operations,
# assuming that 0-axis is batch axis, they get only 1 value instead of batch_size values.
# It could lead to inaccurate/incorrect statistics result.
OPERATIONS_OUTPUT_HAS_NO_BATCH_AXIS = [
    ov_metatypes.OVSpaceToBatchMetatype,
    ov_metatypes.OVBatchToSpaceMetatype,
    ov_metatypes.OVROIPoolingMetatype,
    ov_metatypes.OVROIAlignMetatype,
    ov_metatypes.OVEmbeddingMetatype,
    ov_metatypes.OVIfMetatype,
]
